// Festlegung der Quarzfrequenz
#ifndef F_CPU                   // optional definieren
#define F_CPU 16000000UL
#endif

// Include von Header-Dateien
#include <stdbool.h>			// Bibliothek fuer Bit-Variable
#include <avr/io.h>				// I/O-Konfiguration (intern weitere Dateien)
#include <avr/interrupt.h>		// Definition von Interrupts
#include <util/delay.h>			// Definition von Delays (Wartezeiten)

// Makros
#define SET_BIT(PORT, BIT)  ((PORT) |=  (1 << (BIT))) // Port-Bit Zustand setzen
#define CLR_BIT(PORT, BIT)  ((PORT) &= ~(1 << (BIT))) // Port-Bit Zustand loeschen
#define TGL_BIT(PORT, BIT)  ((PORT) ^=  (1 << (BIT))) // Port-Bit Zustand wechseln (toggle)

//Multiplexer und Ports
#define MUX_DDR1 DDRB
#define MUX_DDR2 DDRC
#define MUX_DDR3 DDRD

//Konstanten
#define N_SERVOS    18

#define MASKE_Servo  0b11110000

#define SERVO_MIN   1000 // ms
#define SERVO_MAX   2000 // ms
#define SERVO_MID   (SERVO_MIN + SERVO_MAX) / 2

#define SERVO_FRAME 100000 // ms (10Hz)
#define SERVO_TIME_DIV (SERVO_FRAME / N_SERVOS) // Zeitfenster pro Servo

#define SERVOTIMER(us) ((us) * (uint16_t)(F_CPU / 1E6))	// Servozeiten - als Timer1-Ticks einzugeben (mit SERVOTIMER).
volatile uint16_t servoTime[N_SERVOS];					// Nur mit Interrupts deaktiviert vernderbar

// Funktionen =================================================================
void servoStart(void)
{
	MUX_DDR1 |= 0xf; //Pins auf Ausgang
	MUX_DDR2 |= 0xf;
	MUX_DDR3 |= 0xf;
	// Setup a first compare match
	TCNT1=0; //TCNT=Timer/Counter1 Register Zhler
	OCR1A = SERVOTIMER(100); //Output Compare Register
	TCCR1B = (1 << CS10);		// Timer/Counter1 Control Register Starte Timer 1 ohne Prescaler
	TIMSK1 |= (1 << OCIE1A);	// Compare Interrupt erlauben Wenn Interrupt dann Timer1_Comp_vect
}

void servoSet(uint8_t servo, uint16_t time /* microseconds */)
{
	switch (time)
	{
		case 0:
		time=1000;
		break;
		case 45:
		time=(SERVO_MID + SERVO_MIN) / 2;
		break;
		case 90:
		time=SERVO_MID;
		break;
		case 135:
		time=(SERVO_MID + SERVO_MAX) / 2;
		break;
		case 180:
		time=SERVO_MAX;
		break;
	}

	uint16_t ticks = SERVOTIMER(time);
	
	cli();
	servoTime[servo] = ticks;
	sei();	// Global Interrupts aktivieren
}

ISR(TIMER1_COMPA_vect)

{
	static uint16_t nextStart= 0;
	static uint8_t servoNr= 0;
	static uint8_t servoOutNr= 0;
	static uint8_t portOutNr= 0;
	static bool outputHigh= 1;
	uint16_t currentTime= OCR1A;
	uint8_t portContent= 0;
	uint8_t portRegister= 0;

	portRegister= 0x05 + 3* portOutNr;
	portContent = _SFR_IO8(portRegister);
	portContent&= MASKE_Servo;					//MUX wird konfiguriert
	portContent+= servoOutNr + (outputHigh<<3);
	_SFR_IO8(portRegister)=portContent;

	if (outputHigh){
		OCR1A    = currentTime + servoTime[servoNr];                 //Endzeit mit derzeitiger zeit berechneter fr Servo
		nextStart = currentTime + SERVOTIMER(SERVO_TIME_DIV);		//Nchster Start
		} else {
		servoNr++;
		if    (++servoOutNr >= 6) {				//berprfung ob am letzten Servo des MUX angelangt
			servoOutNr = 0 ;
			if (++portOutNr >= 3) {			//berprfung ob am letzten MUX angelangt
				portOutNr  = 0 ;
				servoNr    = 0;
			};
		};
		OCR1A = nextStart;					//Nchster Start wird auf OCR geschrieben
	};
	outputHigh = 1-outputHigh;
}